# coding=utf-8
# ======================================
# File:     test_eastmoney.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2023-11-01
# Desc:
#   Unittest for all eastmoney data
# acquiring functions and apis.
# ======================================

import unittest

import pandas as pd

from qteasy.emfuncs import acquire_data, get_k_history


class TestEastmoney(unittest.TestCase):
    """ Test eastmoney data acquiring functions and apis """

    def SetUp(self):
        pass

    def test_get_k_history(self):
        """ Test get_k_history function """
        code = '000651'
        res = get_k_history(code=code, beg='20231102', klt=5)
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        code = '000651'
        res = get_k_history(code=code, beg='20231102', klt=101, verbose=True)
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        code = '000651.SZ'
        res = get_k_history(code=code, beg='20231102', klt=101, verbose=True)
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        code = '000001.SH'
        res = get_k_history(code=code, beg='20231102', klt=101, verbose=True)
        print(res)
        self.assertIsInstance(res, pd.DataFrame)

    def test_stock_live_daily_price(self):
        """ Test stock_live_kline_price function """

        code = ['000016.SZ', '000025.SZ', '000333.SZ']
        res = acquire_data('stock_live_kline_price', symbols=code, freq='D')
        print(f'data acquied for codes [\'000016.SZ\', \'000025.SZ\', \'000333.SZ\']: {res}')
        self.assertIsInstance(res, pd.DataFrame)
        self.assertFalse(res.empty)
        self.assertEqual(res.columns.to_list(), ['symbol', 'open', 'close', 'high', 'low', 'vol', 'amount'])
        self.assertEqual(res.index.name, 'trade_time')
        self.assertTrue(all(item in code for item in res.symbol))
        # some items may not have real time price at the moment
        # self.assertTrue(all(item in res.symbol.to_list() for item in code))

        code = ['000016.SZ', '000025.SZ', '000333.SZ']
        res = acquire_data('stock_live_kline_price', symbols=code, freq='D', verbose=True)
        print(res)
        self.assertIsInstance(res, pd.DataFrame)

        code = ['000016.SZ', '000025.SZ', '000333.SZ', '000404.SZ', '000428.SZ', '000521.SZ', '000541.SZ', '000550.SZ',
                '000572.SZ', '000625.SZ', '000651.SZ', '000721.SZ', '000753.SZ', '000757.SZ', '000800.SZ', '000801.SZ',
                '000810.SZ', '000868.SZ', '000921.SZ', '000951.SZ', '000957.SZ', '000996.SZ', '001259.SZ', '002005.SZ',
                '002011.SZ', '002032.SZ', '002035.SZ', '002050.SZ', '002076.SZ', '002186.SZ', '002242.SZ', '002290.SZ',
                '002403.SZ', '002418.SZ', '002420.SZ', '002429.SZ', '002508.SZ', '002543.SZ', '002594.SZ', '002614.SZ',
                '002668.SZ', '002676.SZ', '002677.SZ', '002705.SZ', '002723.SZ', '002758.SZ', '002759.SZ', '002848.SZ',
                '002860.SZ', '002959.SZ', '003023.SZ', '300100.SZ', '300160.SZ', '300217.SZ', '300272.SZ', '300342.SZ',
                '300403.SZ', '300625.SZ', '300808.SZ', '300824.SZ', '300825.SZ', '300894.SZ', '300911.SZ', '301008.SZ',
                '301039.SZ', '301073.SZ', '301187.SZ', '301215.SZ', '301332.SZ', '301525.SZ', '600006.SH', '600060.SH',
                '600066.SH', '600099.SH', '600104.SH', '600166.SH', '600213.SH', '600258.SH', '600261.SH', '600297.SH',
                '600303.SH', '600335.SH', '600336.SH', '600375.SH', '600386.SH', '600418.SH', '600619.SH', '600653.SH',
                '600686.SH', '600690.SH', '600733.SH', '600754.SH', '600822.SH', '600839.SH', '600854.SH', '600983.SH',
                '601007.SH', '601127.SH', '601238.SH', '601258.SH', '601633.SH', '601956.SH', '601965.SH', '603195.SH',
                '603215.SH', '603219.SH', '603303.SH', '603311.SH', '603355.SH', '603366.SH', '603377.SH', '603486.SH',
                '603515.SH', '603519.SH', '603579.SH', '603657.SH', '603677.SH', '603726.SH', '603868.SH', '605108.SH',
                '605336.SH', '605365.SH', '605555.SH', '688169.SH', '688609.SH', '688696.SH', '688793.SH']
        res = acquire_data('stock_live_kline_price', symbols=code, freq='M')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        self.assertFalse(res.empty)
        self.assertEqual(res.columns.to_list(), ['symbol', 'open', 'close', 'high', 'low', 'vol', 'amount'])
        self.assertEqual(res.index.name, 'trade_time')
        self.assertTrue(all(item in code for item in res.symbol))
        # some items may not have real time price at the moment
        # self.assertTrue(all(item in res.symbol.to_list() for item in code))

        print('test acquiring Index prices')
        codes = ['000001.SH', '000300.SH', '399001.SZ']
        res = acquire_data('stock_live_kline_price', symbols=codes, freq='D')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        self.assertFalse(res.empty)
        self.assertEqual(res.columns.to_list(), ['symbol', 'open', 'close', 'high', 'low', 'vol', 'amount'])
        self.assertEqual(res.index.name, 'trade_time')
        self.assertTrue(all(item in codes for item in res.symbol))
        # some items may not have real time price at the moment
        # self.assertTrue(all(item in res.symbol.to_list() for item in code))

        print('test acquiring ETF price data')
        codes = ['510050.SH', '510300.SH', '510500.SH', '510880.SH', '510900.SH', '512000.SH', '512010.SH']
        res = acquire_data('stock_live_kline_price', symbols=codes, freq='D')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        self.assertFalse(res.empty)
        self.assertEqual(res.columns.to_list(), ['symbol', 'open', 'close', 'high', 'low', 'vol', 'amount'])
        self.assertEqual(res.index.name, 'trade_time')
        self.assertTrue(all(item in codes for item in res.symbol))
        # some items may not have real time price at the moment
        # self.assertTrue(all(item in res.symbol.to_list() for item in code))


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
# ======================================
# File:     trade_recording.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2023-04-04
# Desc:
#   functions that create, modify and
# read from system datatables that store
# trading data in live-trade mode.
# ======================================


import pandas as pd
import numpy as np

from qteasy.database import DataSource

# TODO: add TIMEZONE to qt config arguments
# TODO: move all secondary functions to trading_util.py
TIMEZONE = 'Asia/Shanghai'


# TODO: 创建一个模块级变量，用于存储交易信号的数据源，所有的交易信号都从这个数据源中读取
#  避免交易信号从不同的数据源中获取，导致交易信号的不一致性 ?? 这是不是最好的做法？？
# 9 foundational functions for account and position management
def new_account(user_name, cash_amount, data_source=None, **account_data) -> int:
    """ 创建一个新的账户

    Parameters
    ----------
    user_name: str
        用户名
    cash_amount: float
        账户的初始资金
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源
    account_data: dict
        账户的其他信息

    Returns
    -------
    int: 账户的id
    """
    # 输入数据检查在这里进行
    if cash_amount <= 0:
        raise ValueError('cash_amount must be positive!')

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        raise TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')

    account_id = data_source.insert_sys_table_data(
            'sys_op_live_accounts',
            **{
                'user_name': user_name,
                'created_time': pd.to_datetime('today').strftime('%Y-%m-%d %H:%M:%S'),
                'cash_amount': cash_amount,
                'available_cash': cash_amount,
                'total_invest': cash_amount,
            },
            **account_data,
    )
    return int(account_id)


def get_all_accounts(data_source=None) -> pd.DataFrame:
    """ 获取所有的账户

    Parameters
    ----------
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    accounts: pandas.DataFrame
        所有账户的信息
    """

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        raise TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')

    accounts = data_source.read_sys_table_data('sys_op_live_accounts')

    return accounts


def get_account(account_id, user_name=None, data_source=None) -> dict:
    """ 根据account_id, 或者user_name获取账户的信息

    Parameters
    ----------
    account_id: int
        账户的id
    user_name: str, optional
        用户名, 默认为None, 表示使用account_id, 如果给出了user_name, 则忽略account_id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    dict: ID为account_id的账户的信息

    Raises
    ------
    KeyError: 如果账户不存在，则抛出异常
    """

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        raise TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
    if user_name is None:
        account = data_source.read_sys_table_record('sys_op_live_accounts', record_id=account_id)
        if account == {}:
            raise KeyError(f'Account (id={account_id}) not found!')
        return account
    else:
        account = data_source.read_sys_table_data('sys_op_live_accounts', user_name=user_name)
        if account.empty:
            raise KeyError(f'Account (user_name={user_name}) not found!')
        return account.to_dict(orient='records')[0]


def update_account(account_id, data_source=None, **account_data) -> None:
    """ 更新账户信息

    通用接口，用于更新账户的所有信息，除了账户的持仓和可用资金
    以外，还包括账户的其他状态变量

    Parameters
    ----------
    account_id: int
        账户的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源
    account_data: dict
        交易结果

    Returns
    -------
    None
    """

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        raise TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')

    data_source.update_sys_table_data('sys_op_live_accounts', record_id=account_id, **account_data)


def update_account_balance(account_id, data_source=None, **cash_change) -> None:
    """ 更新账户的资金总额和可用资金, 为了避免误操作，仅允许修改现金总额、可用现金和总投资额，其他字段不可修改

    Parameters
    ----------
    account_id: int
        账户的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源
    cash_change: dict, optional {'cash_amount_change': float,
                                 'available_cash_change': float,
                                 'total_investment_change': float}
        可用资金的变化，其余字段不可用此函数修改

    Returns
    -------
    None
    """
    # TODO: refract this function, allow only one cash value as input argument for cash change, and
    #  change available cash and total investment according to two more input arguments:
    #  update_available_cash: bool default False and update_total_investment: bool default False

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        raise TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')

    account_data = data_source.read_sys_table_record('sys_op_live_accounts', record_id=account_id)
    if account_data == {}:
        raise RuntimeError(f'Account not found! account id: {account_id}')

    cash_amount_change = cash_change.get('cash_amount_change', 0.0)
    if not isinstance(cash_amount_change, (int, float, np.int64, np.float64)):
        raise TypeError(f'cash_amount_change must be a number, got {type(cash_amount_change)} instead')
    cash_amount = account_data['cash_amount'] + cash_amount_change

    available_cash_change = cash_change.get('available_cash_change', 0.0)
    if not isinstance(available_cash_change, (int, float, np.int64, np.float64)):
        raise TypeError(f'available_cash_change must be a number, got {type(available_cash_change)} instead')
    available_cash = account_data['available_cash'] + available_cash_change

    total_investment_change = cash_change.get('total_investment_change', 0.0)
    if not isinstance(total_investment_change, (int, float, np.int64, np.float64)):
        raise TypeError(f'total_investment_change must be a number, got {type(total_investment_change)} instead')
    total_investment = account_data['total_invest'] + total_investment_change

    # 如果可用现金超过现金总额，则报错
    if available_cash > cash_amount:
        raise RuntimeError(f'available_cash ({available_cash}) cannot be greater than cash_amount ({cash_amount})')
    # 如果可用现金小于0，则报错
    if available_cash < 0:
        raise RuntimeError(f'available_cash ({available_cash}) cannot be less than 0!')
    # 如果现金总额小于0，则报错
    if cash_amount < 0:
        raise RuntimeError(f'cash_amount cannot be less than 0!')

    # 更新账户的资金总额和可用资金
    data_source.update_sys_table_data(
            'sys_op_live_accounts',
            record_id=account_id,
            cash_amount=cash_amount,
            available_cash=available_cash,
            total_invest=total_investment,
    )


def delete_account(account_id: int, data_source=None, keep_account_id=True) -> None:
    """ 删除账户，删除账户时，需要同时删除账户的持仓和交易信号。如果账户不存在，则直接返回None
    在删除账户时，同时删除账户的持仓和交易信号，如果删除账户的持仓或交易信号失败，则会回滚删除操作

    Parameters
    ----------
    account_id: int
        账户的id
    data_source: DataSource, optional
        进行操作的数据源, 默认为None, 表示使用默认的数据源
    keep_account_id: bool, optional, default True
        是否保留account_id, 如果为False, 则删除account_id, 该id不再使用，如果为True, 则保留account_id

    Returns
    -------
    None
    """

    if data_source is None:
        import qteasy as qt
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, DataSource):
        msg = f'data_source must be a DataSource instance, got {type(data_source)} instead'
        raise TypeError(msg)

    # 检查account_id是否存在，如果不存在，则直接返回None
    account = get_account(account_id, data_source=data_source)
    if account is None:
        # Nothing is deleted
        return None

    # 找出与account_id相关的所有持仓和交易订单记忆交易结果
    positions = get_account_positions(account_id, data_source=data_source)
    pos_ids = positions.index.tolist() if not positions.empty else []
    orders = query_trade_orders(account_id=account_id, data_source=data_source)
    order_ids = orders.index.tolist() if not orders.empty else []
    results = read_trade_results_by_order_id(order_id=order_ids, data_source=data_source)
    result_ids = results.index.tolist() if not results.empty else []

    # 临时保留将要操作的完整数据表，以便回滚
    account_table_data = data_source.read_sys_table_data('sys_op_live_accounts')
    position_table_data = data_source.read_sys_table_data('sys_op_positions')
    order_table_data = data_source.read_sys_table_data('sys_op_trade_orders')
    result_table_data = data_source.read_sys_table_data('sys_op_trade_results')

    # 开始删除数据：
    print(f'Deleting account {account_id} ...')
    print(f'Following records will be deleted: \n'
          f'account: {account_id}\n'
          f'positions: {pos_ids}\n'
          f'orders: {order_ids}\n'
          f'results: {result_ids}')
    try:
        if not keep_account_id:
            # 删除账户
            data_source.delete_sys_table_data('sys_op_live_accounts', record_ids=[account_id])
        else:
            # 保留账户，但是删除账户的持仓并重置创建日期
            refreshed_data = {'created_time': pd.to_datetime('today').strftime('%Y-%m-%d %H:%M:%S'),
                              'cash_amount': 0,
                              'available_cash': 0,
                              'total_invest': 0,}
            data_source.update_sys_table_data('sys_op_live_accounts',
                                              record_id=account_id,
                                              **refreshed_data)
        # 删除持仓
        data_source.delete_sys_table_data('sys_op_positions', record_ids=pos_ids)
        # 删除交易信号
        data_source.delete_sys_table_data('sys_op_trade_orders', record_ids=order_ids)
        # 删除交易结果
        data_source.delete_sys_table_data('sys_op_trade_results', record_ids=result_ids)
    except Exception as e:
        # 如果删除失败，则回滚删除操作
        data_source.drop_table_data('sys_op_live_accounts')
        data_source.drop_table_data('sys_op_positions')
        data_source.drop_table_data('sys_op_trade_orders')
        data_source.drop_table_data('sys_op_trade_results')
        data_source.write_table_data(account_table_data, 'sys_op_live_accounts')
        data_source.write_table_data(position_table_data, 'sys_op_positions')
        data_source.write_table_data(order_table_data, 'sys_op_trade_orders')
        data_source.write_table_data(result_table_data, 'sys_op_trade_results')
        import warnings
        msg = f'Error occurred: {e}, delete account failed, rollback to previous state!'
        warnings.warn(msg, RuntimeWarning)


def get_position_by_id(pos_id, data_source=None):
    """ 通过pos_id获取持仓的信息

    Parameters
    ----------
    pos_id: int
        持仓的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    持仓的信息
    dict: {'account_id': int,
           'symbol': str,
           'position': str,
           'qty': float,
           'available_qty': float,
           'cost': float,
           }

    Raises
    ------
    RuntimeError: 如果持仓不存在，则报错
    """

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        raise TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')

    position = data_source.read_sys_table_record('sys_op_positions', record_id=pos_id)
    if position == {}:
        raise RuntimeError('Position not found!')
    return position


def get_position_ids(account_id, symbol=None, position_type=None, data_source=None):
    """ 根据symbol和position_type获取账户的持仓id, 如果没有持仓，则返回空列表, 如果有多个持仓，则返回所有持仓的id

    Parameters
    ----------
    account_id: int
        账户的id
    symbol: str, optional
        交易标的的代码，一次只能提供一个symbol
    position_type: str, optional, {'long', 'short'}
        持仓类型, 'long' 表示多头持仓, 'short' 表示空头持仓
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    position_ids: list of int: 持仓的id列表, 如果没有持仓, 则返回空列表
    """

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        raise TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')

    # 获取账户的持仓
    position_filter = {'account_id': account_id}
    if symbol is not None:
        position_filter['symbol'] = str(symbol)
    if position_type is not None:
        position_filter['position'] = str(position_type)

    try:
        position = data_source.read_sys_table_data(
                table='sys_op_positions',
                **position_filter,
        )
    except Exception as e:
        print(f'Error occurred: {e}')
        return []
    if position is None:
        return []
    return position.index.tolist()


def get_or_create_position(account_id: int, symbol: str, position_type: str, data_source: DataSource = None):
    """ 获取账户的持仓, 如果持仓不存在，则创建一条新的持仓记录

    Parameters
    ----------
    account_id: int
        账户的id
    symbol: str
        交易标的的代码
    position_type: str, {'long', 'short'}
        持仓类型, 'long'表示多头持仓, 'short'表示空头持仓
    data_source: DataSource, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    int: 返回持仓记录的id，如果匹配的持仓记录不存在，则创建一条新的空持仓记录，并返回新持仓记录的id
    """

    from qteasy import DataSource, QT_DATA_SOURCE
    if data_source is None:
        data_source = QT_DATA_SOURCE
    if not isinstance(data_source, DataSource):
        raise TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')

    # 检查account_id是否存在，如果不存在，则报错，否则创建的持仓记录将无法关联到账户
    account = get_account(account_id, data_source=data_source)
    if account is None:
        raise RuntimeError(f'account_id {account_id} not found!')

    if not isinstance(symbol, str):
        raise TypeError(f'symbol must be a str, got {type(symbol)} instead')
    if not isinstance(position_type, str):
        raise TypeError(f'position_type must be a str, got {type(position_type)} instead')
    if position_type not in ('long', 'short'):
        raise ValueError(f'position_type must be "long" or "short", got {position_type} instead')
    # debug
    # print(f'[DEBUG]account_id: {account_id}, symbol: {symbol}, position_type: {position_type}')
    position = data_source.read_sys_table_data(
            table='sys_op_positions',
            **{
                'account_id': account_id,
                'symbol': symbol,
                'position': position_type,
            },
    )
    if position.empty:
        return data_source.insert_sys_table_data(
                table='sys_op_positions',
                **{
                    'account_id':account_id,
                    'symbol': symbol,
                    'position': position_type,
                    'qty': 0,
                    'available_qty': 0,
                    'cost': 0,
                },
        )
    # position已存在，此时position中只能有一条记录，否则说明记录重复，报错
    if len(position) > 1:
        raise RuntimeError(f'position record is duplicated, got {len(position)} records: \n{position}')
    return position.index[0]


def update_position(position_id, data_source=None, **position_data):
    """ 更新账户的持仓，包括持仓的数量和可用数量，account_id, position和symbol不可修改

    Parameters
    ----------
    position_id: int
        持仓的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源
    position_data: dict, optional, {'qty_change': float, 'available_qty_change': float, 'cost': float}
        持仓的数据，只能修改qty, available_qty, cost 这三类数据中的任意一个或多个

    Returns
    -------
    None
    """
    # TODO: (maybe) refract this function, require only one change value as argument, change
    #       the available qty according to boolean argument 'change_available_qty', and change
    #       the cost automatically according to the average cost of increased qty

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    if not isinstance(position_id, (int, np.int64)):
        err = TypeError(f'position_id must be an int, got {type(position_id)} instead')
        raise err

    # 从数据库中读取持仓数据，修改后再写入数据库
    position = data_source.read_sys_table_record('sys_op_positions', record_id=position_id)
    if position == {}:
        err = RuntimeError(f'position_id {position_id} not found!')
        raise err

    qty_change = position_data.get('qty_change', 0.0)
    if not isinstance(qty_change, (int, float, np.int64, np.float64)):
        err = TypeError(f'qty_change must be a int or float, got {type(qty_change)} instead')
        raise err
    qty = position['qty'] + qty_change

    available_qty_change = position_data.get('available_qty_change', 0.0)
    if not isinstance(available_qty_change, (int, float, np.int64, np.float64)):
        err = TypeError(f'available_qty_change must be a int or float, got {type(available_qty_change)} instead')
        raise err
    available_qty = position['available_qty'] + available_qty_change

    prev_cost = position['cost']
    cost = position_data.get('cost', prev_cost)
    if cost is not None:
        if not isinstance(cost, (int, float, np.int64, np.float64)):
            err = TypeError(f'cost must be a int or float, got {type(cost)} instead')
            raise err

    # 如果可用数量超过持仓数量，则报错
    if available_qty > qty:
        err = RuntimeError(f'available_qty ({position["available_qty"]}+{available_qty_change}={available_qty})'
                           f' cannot be greater than qty ({position["qty"]}+{qty_change}={qty})')
        raise err
    # 如果可用数量小于0，则报错
    if available_qty < 0:
        err = RuntimeError(f'available_qty ({position["available_qty"]}+{available_qty_change}={available_qty})'
                           f' cannot be less than 0!')
        raise err
    # 如果持仓数量小于0，则报错
    if qty < 0:
        err = RuntimeError(f'qty ({position["qty"]}+{qty_change}={qty}) cannot be less than 0!')
        raise err

    data_source.update_sys_table_data(
            'sys_op_positions',
            record_id=position_id,
            qty=qty,
            available_qty=available_qty,
            cost=cost,
    )


def get_account_positions(account_id, data_source=None):
    """ 根据account_id获取账户的所有持仓

    Parameters
    ----------
    account_id: int
        账户的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    positions: pandas.DataFrame, columns=['account_id', 'symbol', 'position', 'qty', 'available_qty']
        账户的所有持仓
        account_id: int, 账户的id
        symbol: str, 股票代码
        position: str, 持仓类型，'long'表示多头持仓，'short'表示空头持仓
        qty: int, 持仓数量
        available_qty: int, 可用数量
    None: 如果账户不存在或持仓不存在，则返回None
    """

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    positions = data_source.read_sys_table_data(
            'sys_op_positions',
            **{'account_id': account_id},
    )
    if positions is None:
        return pd.DataFrame(columns=['account_id', 'symbol', 'position', 'qty', 'available_qty'])
    return positions


# Four 2nd foundational functions for account and position availability check
def get_account_cash_availabilities(account_id, data_source=None):
    """ 根据账户id获取账户的可用资金和资金总额
    返回一个tuple，第一个元素是账户的可用资金，第二个元素是账户的资金总额

    Parameters
    ----------
    account_id: int
        账户的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    cash_availabilities: tuple
        (cash_amount: float, 账户的可用资金
         available_cash: float, 账户的资金总额
         total_invest: float, 账户的总投入
        )
    """

    account = get_account(account_id=account_id, data_source=data_source)
    return account['cash_amount'], account['available_cash'], account['total_invest']


def get_account_position_availabilities(account_id, shares=None, data_source=None):
    """ 根据account_id读取账户的持仓，筛选出与shares相同的symbol的持仓，返回两个ndarray，分别为
    每一个share对应的持仓的数量和可用数量

    Parameters
    ----------
    account_id: int
        账户的id
    shares: list of str, optional
        需要输出的持仓的symbol列表, 如果不给出shares，则返回所有持仓的数量和可用数量
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    (symbols, own_amounts, available_amounts, costs): tuple, (list, ndarray, ndarray, ndarray)
        symbols: 持仓share的symbol列表
        own_amounts: 每一个share对应的持仓的数量
        available_amounts: 每一个share对应的持仓的可用数量
        costs: 每一个share对应的持仓的成本
    """

    # 根据account_id读取账户的全部持仓
    positions = get_account_positions(account_id=account_id, data_source=data_source)

    if positions is None:
        return shares, np.zeros(len(shares)), np.zeros(len(shares))
    # 如果没有给出shares，则读取账户中所有持仓的symbol
    if shares is None:
        shares = positions['symbol'].unique()
    if isinstance(shares, str):
        from qteasy.utilfuncs import str_to_list
        shares = str_to_list(shares)
    if not isinstance(shares, (list, tuple, np.ndarray)):
        err = TypeError(f'shares must be a list, tuple or ndarray, got {type(shares)} instead')
        raise err

    own_amounts = []
    available_amounts = []
    costs = []
    for share in shares:
        # 检查symbol为share的持仓是否存在
        position = positions[(positions['symbol'] == share) & (positions['qty'] > 0)]
        if position.empty:
            own_amounts.append(0.0)
            available_amounts.append(0.0)
            costs.append(0.0)
            continue
        # 如果同时存在多头和空头持仓，则报错
        if len(position) > 1:
            err = RuntimeError(f'position for {share} has more than one position!')
            raise err
        position = position.iloc[0]
        # 如果存在多头持仓，则将多头持仓的数量和可用数量放入列表
        if position['position'] == 'long':
            own_amounts.append(position['qty'])
            available_amounts.append(position['available_qty'])
            costs.append(position['cost'])
            continue
        # 如果存在空头持仓，则将空头持仓的数量和可用数量乘以-1并放入列表
        if position['position'] == 'short':
            own_amounts.append(-1 * position['qty'])
            available_amounts.append(-1 * position['available_qty'])
            costs.append(position['cost'])
            continue
    # 如果列表长度与shares长度不相等，则报错
    if len(own_amounts) != len(shares):
        err = RuntimeError(f'own_amounts length ({len(own_amounts)}) is not equal to shares length ({len(shares)})')
        raise err
    if len(available_amounts) != len(shares):
        err = RuntimeError(f'available_amounts length ({len(available_amounts)}) is not equal to '
                           f'shares length ({len(shares)})')
        raise err
    if len(costs) != len(shares):
        err = RuntimeError(f'costs length ({len(costs)}) is not equal to shares length ({len(shares)})')
        raise err
    # 将列表转换为ndarray并返回
    result = (
        shares,
        np.array(own_amounts).astype('float'),
        np.array(available_amounts).astype('float'),
        np.array(costs).astype('float')
    )

    return result


def get_account_position_details(account_id, shares=None, data_source=None):
    """ 根据account_id读取账户的持仓，筛选出与shares相同的symbol的持仓，返回一个DataFrame，包含
    每一个share对应的持仓的symbol，position，qty和可用数量。

    Parameters
    ----------
    account_id: int
        账户的id
    shares: list of str, optional
        需要输出的持仓的symbol列表, 如果不给出shares，则返回所有持仓的数量和可用数量
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    positions: DataFrame, columns=['symbol', 'qty', 'available_qty'， 'cost']
        account持仓的symbol，qty, available_qty和cost, symbol与shares的顺序一致
    """

    # 根据account_id读取账户的全部持仓
    symbols, amounts, available_amounts, costs = get_account_position_availabilities(
            account_id=account_id,
            shares=shares,
            data_source=data_source
    )
    # 将symbol，position，qty和available_qty放入DataFrame并返回
    positions = pd.DataFrame(
            {
                'qty': amounts,
                'available_qty': available_amounts,
                'cost': costs,
            },
            index=symbols,
    ).T
    return positions


# 4 foundational functions for trade order
def record_trade_order(order, data_source=None):
    """ 将交易信号写入数据库

    Parameters
    ----------
    order: dict
        标准形式的交易订单，格式为
        {
            'pos_id': int,  # 持仓的id
            'direction': str,  # 交易方向，'buy'或'sell'
            'order_type': str,  # 订单类型，'market'或'limit'
            'qty': float,  # 交易数量
            'price': float,  # 交易价格
            'submitted_time': datetime.datetime,  # 交易信号提交时间
            'status': str,  # 交易信号状态
        }
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    order_id: int
    写入数据库的交易信号的id
    """

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE

    err = None

    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')

    # 检查交易信号的格式和数据合法性
    if not isinstance(order, dict):
        err = TypeError(f'signal must be a dict, got {type(order)} instead')
    if not isinstance(order['pos_id'], (int, np.int64)):
        err = TypeError(f'signal["pos_id"] must be an int, got {type(order["pos_id"])} instead')
    if not isinstance(order['direction'], str):
        err = TypeError(f'signal["direction"] must be a str, got {type(order["direction"])} instead')
    if not isinstance(order['order_type'], str):
        err = TypeError(f'signal["order_type"] must be a str, got {type(order["order_type"])} instead')
    if not isinstance(order['qty'], (float, int, np.float64, np.int64)):
        err = TypeError(f'signal["qty"] must be a float, got {type(order["qty"])} instead')
    if not isinstance(order['price'], (float, int, np.float64, np.int64)):
        err = TypeError(f'signal["price"] must be a float, got {type(order["price"])} instead')
    if order['qty'] <= 0:
        err = RuntimeError(f'signal["qty"] ({order["qty"]}) must be greater than 0!')
    if order['price'] <= 0:
        err = RuntimeError(f'signal["price"] ({order["price"]}) must be greater than 0!')

    if err is not None:
        raise err

    return data_source.insert_sys_table_data('sys_op_trade_orders', **order)


def read_trade_order(order_id, data_source=None) -> dict:
    """ 根据order_id从数据库中读取交易信号

    Parameters
    ----------
    order_id: int
        交易信号的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    signal: dict
        交易信号
    """
    if not isinstance(order_id, (int, np.int64)):
        err = TypeError(f'order_id must be an int, got {type(order_id)} instead')
        raise err

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    return data_source.read_sys_table_record('sys_op_trade_orders', record_id=order_id)


def update_trade_order(order_id, data_source=None, status=None, qty=None, raise_if_status_wrong=False):
    """ 更新数据库中trade_signal的状态或其他信，这里只操作trade_signal，不处理交易结果

    trade_order的所有字段中，可以更新字段只有status和qty(qty只能在submit的时候更改，一旦submit之后就不能再更改。
    status的更新遵循下列规律：
    1. 如果status为 'created'，则可以更新为 'submitted', 同时设置'submitted_time';
    2. 如果status为 'submitted'，则可以更新为 'canceled', 'partial-filled' 或 'filled';
    3. 如果status为 'partial-filled'，则可以更新为 'canceled' 或 'filled';
    4. 如果status为 'canceled' 或 'filled'，则不可以再更新.

    Parameters
    ----------
    order_id: int
        交易信号的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源
    status: str, optional
        交易信号的状态, 默认为None, 表示不更新状态
    qty: float, optional
        交易信号的数量, 默认为None, 表示不更新数量
    raise_if_status_wrong: bool, default False
        如果status不符合规则，则抛出RuntimeError, 默认为False, 表示不抛出异常

    Returns
    -------
    order_id: int, 如果更新成功，返回更新后的交易信号的id
    None, 如果更新失败，返回None

    Raises
    ------
    TypeError
        如果data_source不是DataSource的实例，则抛出TypeError
    RuntimeError
        如果trade_signal读取失败，则抛出RuntimeError
        如果status不符合规则，则抛出RuntimeError
    """

    if status is None:
        return None

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    err = None
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
    if status is not None:
        if not isinstance(status, str):
            err = TypeError(f'status must be a str, got {type(status)} instead')
        if status not in ['created', 'submitted', 'canceled', 'partial-filled', 'filled']:
            err = RuntimeError(f'status ({status}) not in [created, submitted, canceled, partial-filled, filled]!')

    if err is not None:
        raise err

    trade_signal = data_source.read_sys_table_record('sys_op_trade_orders', record_id=order_id)

    # 如果trade_signal读取失败，则报错
    if trade_signal is None:
        err = RuntimeError(f'Trade signal (order_id = {order_id}) not found!')
        raise err

    # 如果trade_signal的状态为 'created'，则可以更新为 'submitted'
    if trade_signal['status'] == 'created' and status == 'submitted':
        if qty is not None:
            assert isinstance(qty, (float, int, np.float64, np.int64)), f'qty must be a float, got {type(qty)} instead'
            assert qty >= 0, f'qty ({qty}) must be greater than or equal to 0!'
        else:
            qty = trade_signal['qty']
        submit_time = pd.to_datetime('today').strftime('%Y-%m-%d %H:%M:%S')
        return data_source.update_sys_table_data(
                'sys_op_trade_orders',
                record_id=order_id,
                submitted_time=submit_time,
                status=status,
                qty=qty,
        )
    # 如果trade_signal的状态为 'submitted'，则可以更新为 'canceled', 'partial-filled' 或 'filled'
    if trade_signal['status'] == 'submitted' and status in ['canceled', 'partial-filled', 'filled']:
        return data_source.update_sys_table_data(
                'sys_op_trade_orders',
                order_id,
                status=status
        )
    # 如果trade_signal的状态为 'partial-filled'，则可以更新为 'canceled' 或 'filled'
    if trade_signal['status'] == 'partial-filled' and status in ['canceled', 'filled']:
        return data_source.update_sys_table_data(
                'sys_op_trade_orders',
                order_id,
                status=status
        )

    if raise_if_status_wrong:
        err = RuntimeError(f'Wrong status update: {trade_signal["status"]} -> {status}')
        raise err

    return


def query_trade_orders(account_id,
                       symbol=None,
                       position=None,
                       direction=None,
                       order_type=None,
                       status=None,
                       data_source=None) -> pd.DataFrame:
    """ 根据symbol、direction、status 从数据库中查询交易信号并批量返回结果

    Parameters
    ----------
    account_id: int
        账户的id
    symbol: str, optional
        交易标的
    position: str, optional, {'long', 'short'}
        持仓方向, 默认为None, 表示不限制, 'long' 表示多头, 'short' 表示空头
    direction: str, optional, {'buy', 'sell'}
        交易方向, 默认为None, 表示不限制, 'buy' 表示买入, 'sell' 表示卖出
    order_type: str, optional, {'market', 'limit', 'stop', 'stop_limit'}
        交易类型, 默认为None, 表示不限制, 'market' 表示市价单, 'limit' 表示限价单, 'stop' 表示止损单, 'stop_limit' 表示止损限价单
    status: str, optional, {'created', 'submitted', 'canceled', 'partial-filled', 'filled'}
        交易信号状态
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    pd.DataFrame, columns = ['pos_id', 'direction', 'order_type', 'qty', 'price', 'submitted_time', 'status']
        交易订单列表，包含所有符合查询条件的订单的DataFrame
    """

    # TODO: move this function to trading_util.py
    # TODO: allow list of str as input for symbol, position, direction, order_type, status

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    # 从数据库中读取position的id
    pos_ids = get_position_ids(account_id, symbol, position, data_source=data_source)
    if pos_ids is None:
        return pd.DataFrame(columns=['pos_id', 'direction', 'order_type', 'qty', 'price', 'submitted_time', 'status'])

    data_filter = {}
    if direction is not None:
        data_filter['direction'] = direction
    if order_type is not None:
        data_filter['order_type'] = order_type
    if status is not None:
        data_filter['status'] = status

    res = []
    for pos_id in pos_ids:
        res.append(
                data_source.read_sys_table_data(
                        'sys_op_trade_orders',
                        pos_id=pos_id,
                        **data_filter,
                )
        )
    if all(r is None for r in res):
        return pd.DataFrame(columns=['pos_id', 'direction', 'order_type', 'qty', 'price', 'submitted_time', 'status'])

    return pd.concat(res).sort_index()


# 2 2nd level functions for trade signal TODO: (maybe) move to trading_util.py
def read_trade_order_detail(order_id, data_source=None) -> dict:
    """ 从数据库中读取交易信号的详细信息，除了trade_order表中已有的信息以外，
     返回值还包括从关联表中读取的信息，包括三个字段：
     account_id, symbol, position

    Parameters
    ----------
    order_id: int
        交易信号的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    trade_order_detail: dict
        包含symbol和position信息的交易信号明细:
        {
            'order_id': int, 订单ID
            'account_id': int,  账户ID
            'pos_id': int,  持仓ID
            'symbol': str,  持仓股票代码
            'position': str,  多头/空头
            'direction': str,  买入/卖出
            'order_type': str,  订单类型
            'qty': float,  挂单数量
            'price': float,  挂单价格
            'submitted_time': str,  订单提交时间
            'status': str,  订单状态
        }
    """

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    trade_order_detail = read_trade_order(order_id, data_source=data_source)
    if trade_order_detail == {}:
        return None
    pos_id = trade_order_detail['pos_id']
    position = get_position_by_id(pos_id, data_source=data_source)
    if position is None:
        err = RuntimeError(f'Position (position_id = {pos_id}) not found!')
        raise err
    # 从关联表中读取symbol和position的信，添加到trade_signal_detail中
    trade_order_detail['order_id'] = order_id
    trade_order_detail['account_id'] = position['account_id']
    trade_order_detail['symbol'] = position['symbol']
    trade_order_detail['position'] = position['position']
    return trade_order_detail


def save_parsed_trade_orders(account_id, symbols, positions, directions, quantities, prices, data_source=None):
    """ 根据parse_trade_signal的结果，将交易订单要素组装成完整的交易订单dict，并将交易信号保存到数据库

    Parameters
    ----------
    account_id: int
        账户ID
    symbols: list of str
        交易信号对应的股票代码
    positions: list of str
        交易信号对应的持仓类型
    directions: list of str
        交易信号对应的交易方向
    quantities: list of float
        交易信号对应的交易数量
    prices: list of float
        交易信号对应的股票价格
    data_source: str, optional
        交易信号对应的数据源, 默认为None, 使用默认数据源

    Returns
    -------
    order_ids: list of int
        生成的交易订单的id
    """

    if len(symbols) == 0:
        return []

    if len(symbols) != len(positions) or \
            len(symbols) != len(directions) or \
            len(symbols) != len(quantities) or \
            len(symbols) != len(prices):
        err = ValueError('Length of symbols, positions, directions, quantities and prices must be the same')
        raise err

    order_ids = []
    # 逐个处理所有的交易信号要素
    for sym, pos, dirc, qty, price in zip(symbols, positions, directions, quantities, prices):
        # 获取pos_id, 如果pos_id不存在，则新建一个posiiton
        pos_id = get_or_create_position(account_id, sym, pos, data_source=data_source)
        # 生成交易信号dict
        trade_order = {
            'pos_id': pos_id,
            'direction': dirc,
            'order_type': 'market',  # TODO: 交易信号的order_type应该是可配置的，增加其他配置选项
            'qty': qty,
            'price': price,
            'submitted_time': None,
            'status': 'created'
        }
        sig_id = record_trade_order(trade_order, data_source=data_source)
        order_ids.append(sig_id)

    return order_ids


# 5 foundational functions for trade result
def write_trade_result(trade_result, data_source=None):
    """ 将交易结果写入数据库, 并返回交易结果的id

    Parameters
    ----------
    trade_result: dict
        交易结果
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    result_id: int
        交易结果的id
    """

    if not isinstance(trade_result, dict):
        err = TypeError('trade_results must be a dict')
        raise err

    if not isinstance(trade_result['order_id'], (int, np.int64)):
        err = TypeError(f'order_id of trade_result must be an int, got {type(trade_result["order_id"])} instead')
        raise err
    if not isinstance(trade_result['filled_qty'], (int, float, np.int64, np.float64)):
        err = TypeError(f'filled_qty of trade_result must be a number, got {type(trade_result["filled_qty"])} instead')
        raise err
    if not isinstance(trade_result['price'], (int, float, np.int64, np.float64)):
        err = TypeError(f'price of trade_result must be a number, got {type(trade_result["price"])} instead')
        raise err
    if not isinstance(trade_result['transaction_fee'], (int, float, np.int64, np.float64)):
        err = TypeError(f'transaction_fee of trade_result must be a number, got '
                        f'{type(trade_result["transaction_fee"])} instead')
        raise err
    if isinstance(trade_result['execution_time'], str):
        try:
            execution_time = pd.to_datetime(trade_result['execution_time']).strftime('%Y-%m-%d %H:%M:%S')
            trade_result['execution_time'] = execution_time
        except Exception as e:
            err = RuntimeError(f'{e}, Invalid execution_time {trade_result["execution_time"]}, '
                               f'can not be converted to datetime format')
            raise err
    if not isinstance(trade_result['canceled_qty'], (int, float, np.int64, np.float64)):
        err = TypeError(f'canceled_qty of trade_result must be a number, got '
                        f'{type(trade_result["canceled_qty"])} instead')
        raise err
    if not isinstance(trade_result['delivery_status'], str):
        err = TypeError(f'delivery_status of trade_result must be a str, '
                        f'got {type(trade_result["delivery_status"])} instead')
        raise err
    if not isinstance(trade_result['delivery_amount'], (int, float, np.int64, np.float64)):
        err = TypeError(f'delivery_amount of trade_result must be a number, got '
                        f'{type(trade_result["delivery_amount"])} instead')
        raise err
    if trade_result['order_id'] <= 0:
        err = ValueError('order_id can not be less than or equal to 0')
        raise err
    if trade_result['filled_qty'] < 0:
        err = ValueError('filled_qty can not be less than 0')
        raise err
    if trade_result['price'] < 0:
        err = ValueError('price can not be less than 0')
        raise err
    if trade_result['transaction_fee'] < 0:
        err = ValueError('transaction_fee can not be less than 0')
        raise err
    if trade_result['canceled_qty'] < 0:
        err = ValueError('canceled_qty can not be less than 0')
        raise err
    if trade_result['delivery_amount'] < 0:
        # raise ValueError('delivery_amount can not be less than 0')
        pass
    if trade_result['delivery_status'] not in ['ND', 'DL']:
        err = ValueError(f'delivery_status can only be ND or DL, got {trade_result["delivery_status"]} instead')
        raise err

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    result_id = data_source.insert_sys_table_data('sys_op_trade_results', **trade_result)
    return result_id


def update_trade_result(result_id, delivery_status, data_source=None):
    """ 更新交易结果的delivery_status

    Parameters
    ----------
    result_id: int
        交易结果的id
    delivery_status: str
        交易结果的delivery_status
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源
    """
    if not isinstance(result_id, (int, np.int64)):
        err = TypeError('result_id must be an int')
        raise err
    if not isinstance(delivery_status, str):
        err = TypeError('delivery_status must be a str')
        raise err
    if delivery_status not in ['ND', 'DL']:
        err = ValueError(f'delivery_status can only be ND or DL, got {delivery_status} instead')
        raise err

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    data_source.update_sys_table_data(
            'sys_op_trade_results',
            result_id,
            **{'delivery_status': delivery_status},
    )


def read_trade_result_by_id(result_id, data_source=None):
    """ 根据result_id从数据库中读取一条交易结果, 并返回交易结果的dict形式

    Parameters
    ----------
    result_id: int
        交易结果的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    trade_result: dict 交易结果
        {'order_id': int,  # 交易结果对应的order_id
        'filled_qty': float,  # 交易结果的成交数量
        'price': float,  # 交易结果的成交价格
        'transaction_fee': float,  # 交易结果的交易费用
        'execution_time': str,  # 交易结果的成交时间
        'canceled_qty': float,  # 交易结果的撤单数量
        'delivery_amount': float,  # 交易结果的交割数量
        'delivery_status': str}  # 交易结果的交割状态
    """
    if not isinstance(result_id, (int, np.int64)):
        err = TypeError('result_id must be an int')
        raise err

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    trade_result = data_source.read_sys_table_record('sys_op_trade_results', record_id=result_id)
    return trade_result


def read_trade_results_by_order_id(order_id, data_source=None):
    """ 根据order_id从数据库中读取所有与signal相关的交易结果，以DataFrame的形式返回

    Parameters
    ----------
    order_id: int, list of int
        交易信号的id, 可以是一个int, 也可以是一个int的list, 表示多个交易信号的id
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    trade_results: pd.DataFrame, columns=['order_id', 'filled_qty', 'price', 'transaction_fee', 'execution_time',
                        'canceled_qty', 'delivery_amount', 'delivery_status']
    交易结果, 包含以下列:
        - order_id: int 交易信号的id
        - filled_qty: int 成交数量
        - price: float 成交价格
        - transaction_fee: float 交易费用
        - execution_time: datetime.datetime 成交时间
        - canceled_qty: int 撤单数量
        - delivery_amount: float 交割金额
        - delivery_status: str 交割状态
    """
    if not isinstance(order_id, (int, np.int64, np.ndarray, list, )):
        err = TypeError(f'order_id must be an integer, a list of integers or a numpy array of integers, '
                        f'got {type(order_id)} instead')
        raise err
    if isinstance(order_id, np.ndarray):
        order_id = order_id.tolist()
    if isinstance(order_id, list):
        if not all([isinstance(i, (int, np.int64)) for i in order_id]):
            err = TypeError('order_id must be an int or a list of int')
            raise err

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    if isinstance(order_id, (int, np.int64)):
        trade_results = data_source.read_sys_table_data(
                'sys_op_trade_results',
                **{'order_id': order_id},
        )
    else:  # order_id is a list
        trade_results = data_source.read_sys_table_data(
            'sys_op_trade_results',
        )
    if trade_results.empty:
        return pd.DataFrame(columns=['order_id', 'filled_qty', 'price', 'transaction_fee', 'execution_time',
                                     'canceled_qty', 'delivery_amount', 'delivery_status'])
    if isinstance(order_id, list):
        trade_results = trade_results[trade_results['order_id'].isin(order_id)]

    return trade_results


def read_trade_results_by_delivery_status(delivery_status, data_source=None) -> pd.DataFrame:
    """ 根据delivery_status从数据库中读取所有与signal相关的交易结果，以DataFrame的形式返回

    Parameters
    ----------
    delivery_status: str
        交易结果的交割状态
    data_source: str, optional
        数据源的名称, 默认为None, 表示使用默认的数据源

    Returns
    -------
    trade_results: pd.DataFrame
        交易结果
    """
    if not isinstance(delivery_status, str):
        err = TypeError('delivery_status must be a str')
        raise err
    if not delivery_status in ['ND', 'DL']:
        err = ValueError(f'delivery_status can only be ND or DL, got {delivery_status} instead')
        raise err

    import qteasy as qt
    if data_source is None:
        data_source = qt.QT_DATA_SOURCE
    if not isinstance(data_source, qt.DataSource):
        err = TypeError(f'data_source must be a DataSource instance, got {type(data_source)} instead')
        raise err

    trade_results = data_source.read_sys_table_data(
            'sys_op_trade_results',
            **{'delivery_status': delivery_status},
    )
    return trade_results



# coding=utf-8
# ======================================
# File:     test_user_defined.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-12
# Desc:
#   Test file for user-defined
# strategies.
# ======================================
import unittest

import qteasy as qt
import numpy as np


def market_value_weighted(stock_return, mv, mv_cat, bp_cat, mv_target, bp_target):
    """ 根据mv_target和bp_target计算市值加权收益率

    """
    sel = (mv_cat == mv_target) & (bp_cat == bp_target)
    mv_total = np.nansum(mv[sel])
    mv_weight = mv / mv_total
    return_total = np.nansum(stock_return[sel] * mv_weight[sel])
    return return_total


class MultiFactors(qt.FactorSorter):

    def __init__(self, pars: tuple = (0.5, 0.3, 0.7)):
        super().__init__(
                pars=pars,
                par_count=3,
                par_types=['float', 'float', 'float'],  # 参数1:大小市值分类界限，参数2:小/中bp分界线，参数3，中/大bp分界线
                par_range=[(0.01, 0.99), (0.01, 0.49), (0.50, 0.99)],
                name='MultiFactor',
                description='根据Fama-French三因子回归模型估算HS300成分股的alpha值选股',
                strategy_run_timing='close',  # 在周期结束（收盘）时运行
                strategy_run_freq='m',  # 每月执行一次选股（每周或每天都可以）
                strategy_data_types='pb, total_mv, close',  # 执行选股需要用到的股票数据
                data_freq='d',  # 数据频率（包括股票数据和参考数据）
                window_length=20,
                use_latest_data_cycle=True,
                reference_data_types='close-000300.SH',  # 选股需要用到市场收益率，作为参考数据传入
                max_sel_count=10,  # 最多选出10支股票
                sort_ascending=True,  # 选择因子最小的股票
                condition='less',  # 仅选择因子小于某个值的股票
                lbound=0,  # 仅选择因子小于0的股票
                ubound=0,  # 仅选择因子小于0的股票
        )

    def realize(self, h, r=None, t=None, pars=None):

        size_gate_percentile, bp_small_percentile, bp_large_percentile = self.pars
        # 读取投资组合的数据PB和total_MV的最新值
        pb = h[:, -1, 0]  # 当前所有股票的PB值
        mv = h[:, -1, 1]  # 当前所有股票的市值
        pre_close = h[:, -2, 2]  # 当前所有股票的前收盘价
        close = h[:, -1, 2]  # 当前所有股票的最新收盘价

        # 读取参考数据(r)
        market_pre_close = r[-2, 0]  # HS300的昨收价
        market_close = r[-1, 0]  # HS300的收盘价

        # 计算账面市值比，为pb的倒数
        bp = pb ** -1
        # 计算市值的50%的分位点,用于后面的分类
        size_gate = np.nanquantile(mv, size_gate_percentile)
        # 计算账面市值比的30%和70%分位点,用于后面的分类
        bm_30_gate = np.nanquantile(bp, bp_small_percentile)
        bm_70_gate = np.nanquantile(bp, bp_large_percentile)
        # 计算每只股票的当日收益率
        stock_return = pre_close / close - 1

        # 根据每只股票的账面市值比和市值，给它们分配bp分类和mv分类
        # 市值小于size_gate的cat为1，否则为2
        mv_cat = np.ones_like(mv)
        mv_cat += (mv > size_gate).astype('float')
        # bp小于30%的cat为1，30%～70%之间为2，大于70%为3
        bp_cat = np.ones_like(bp)
        bp_cat += (bp > bm_30_gate).astype('float')
        bp_cat += (bp > bm_70_gate).astype('float')

        # 获取小市值组合的市值加权组合收益率
        smb_s = (market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 1) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 2) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 3)) / 3
        # 获取大市值组合的市值加权组合收益率
        smb_b = (market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 1) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 2) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 3)) / 3
        smb = smb_s - smb_b
        # 获取大账面市值比组合的市值加权组合收益率
        hml_b = (market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 3) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 3)) / 2
        # 获取小账面市值比组合的市值加权组合收益率
        hml_s = (market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 1) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 1)) / 2
        hml = hml_b - hml_s

        # 计算市场收益率
        market_return = market_pre_close / market_close - 1

        coff_pool = []
        # 对每只股票进行回归获取其alpha值
        for rtn in stock_return:
            x = np.array([[market_return, smb, hml, 1.0]])
            y = np.array([[rtn]])
            # OLS估计系数
            coff = np.linalg.lstsq(x, y)[0][3][0]
            coff_pool.append(coff)

        # 以alpha值为股票组合的选股因子执行选股
        factors = np.array(coff_pool)

        return factors


class IndexEnhancement(qt.GeneralStg):

    def __init__(self, pars: tuple = (0.35, 0.8, 5)):
        super().__init__(
                pars=pars,
                par_count=2,
                par_types=['float', 'float', 'int'],  # 参数1:沪深300指数权重阈值，低于它的股票不被选中，参数2: 初始权重，参数3: 连续涨跌天数，作为强弱势判断阈值
                par_range=[(0.01, 0.99), (0.51, 0.99), (2, 20)],
                name='IndexEnhancement',
                description='跟踪HS300指数选股，并根据连续上涨/下跌趋势判断强弱势以增强权重',
                strategy_run_timing='close',  # 在周期结束（收盘）时运行
                strategy_run_freq='d',  # 每天执行一次选股
                strategy_data_types='wt-000300.SH, close',  # 利用HS300权重设定选股权重, 根据收盘价判断强弱势
                data_freq='d',  # 数据频率（包括股票数据和参考数据）
                window_length=20,
                use_latest_data_cycle=True,
                reference_data_types='',  # 不需要使用参考数据
        )

    def realize(self, h, r=None, t=None, pars=None):
        weight_threshold, init_weight, price_days = self.pars
        # 读取投资组合的权重wt和最近price_days天的收盘价
        wt = h[:, -1, 0]  # 当前所有股票的权重值
        pre_close = h[:, -price_days - 1:-1, 1]
        close = h[:, -price_days:, 1]  # 当前所有股票的最新连续收盘价
        # 计算连续price_days天的收益
        stock_returns = pre_close - close  # 连续p天的收益

        # 设置初始选股权重为0.8
        weights = init_weight * np.ones_like(wt)

        # 剔除掉权重小于weight_threshold的股票
        weights[wt < weight_threshold] = 0

        # 找出强势股，将其权重设为1, 找出弱势股，将其权重设置为 init_weight - (1 - init_weight)
        up_trends = np.all(stock_returns > 0, axis=1)
        weights[up_trends] = 1.0
        down_trend_weight = init_weight - (1 - init_weight)
        down_trends = np.all(stock_returns < 0, axis=1)
        weights[down_trends] = down_trend_weight

        # 实际选股权重为weights * HS300权重
        weights *= wt
        # print(f'select weight is: \n{weights}')

        return weights


class GridTrading(qt.GeneralStg):

    def __init__(self, pars: tuple = (2.0, 3.0, 0.3, 0.5, 300)):
        super().__init__(
                pars=pars,
                par_count=5,
                par_types=['float', 'float', 'float', 'float', 'int'],
                # 仓位配置的阈值：参数1:低仓位阈值，参数2: 高仓位阈值，参数3：低仓位比例，参数4:高仓位比例，参数5:计算天数
                par_range=[(0.5, 3.0), (2.0, 10.), (0.01, 0.5), (0.5, 0.99), (10, 300)],
                name='GridTrading',
                description='根据过去300份钟的股价均值和标准差，改变投资金额的仓位',
                strategy_run_timing='close',  # 在周期结束（收盘）时运行
                strategy_run_freq='1min',  # 每份钟执行一次调整
                strategy_data_types='close',  # 使用份钟收盘价调整
                data_freq='1min',  # 数据频率（包括股票数据和参考数据）
                window_length=300,
                use_latest_data_cycle=False,  # 高频数据不需要使用当前数据区间
                reference_data_types='',  # 不需要使用参考数据
        )

    def realize(self, h, r=None, t=None, pars=None):
        """策略输出PT信号，即仓位目标信号"""

        low_threshold, high_threshold, low_pos, hi_pos, days = self.pars

        # 读取最近N天的收盘价
        close = h[:, - days:, 0]  # 最新连续收盘价
        current_close = h[:, -1, 0]  # 当天的收盘价

        # 计算N天的平均价和标准差，并计算仓位阈值
        close_mean = np.nanmean(close, axis=1)
        close_std = np.nanstd(close, axis=1)
        hi_positive = close_mean + high_threshold * close_std
        low_positive = close_mean + low_threshold * close_std
        low_negative = close_mean - low_threshold * close_std
        hi_negative = close_mean - high_threshold * close_std

        # 根据当前的实际价格确定目标仓位，并将目标仓位作为信号输出
        pos = np.zeros_like(close_mean)
        pos = np.where(current_close > hi_positive, hi_pos, pos)
        pos = np.where(hi_positive >= current_close > low_positive, low_pos, pos)
        pos = np.where(low_positive >= current_close > low_negative, 0, pos)
        pos = np.where(low_negative >= current_close > hi_negative, - low_pos, pos)
        pos = np.where(current_close >= hi_negative, - hi_pos, pos)

        return pos


class FastExperiments(unittest.TestCase):
    """This test case is created to have experiments done that can be quickly called from Command line"""

    def setUp(self):
        pass

    def test_multi_factors(self):
        """ test strategy MultiFactors"""
        self.assertEqual(1, 1)
        shares = qt.filter_stock_codes(index='000300.SH', date='20210101')
        print(len(shares), shares[:10])

        alpha = MultiFactors()
        op = qt.Operator(alpha, signal_type='PT')

        op.op_type = 'stepwise'
        op.set_blender("0.8*s0", 'close')
        # op.run(mode=1,
        #        invest_start='20210101',
        #        invest_end='20220501',
        #        asset_type='E',
        #        invest_cash_amounts=[1000000],
        #        asset_pool=shares,
        #        trade_batch_size=100,
        #        sell_batch_size=1,
        #        trade_log=True)
        self.assertTrue(True)

    def test_index_enhancement(self):
        """ tests for self-defined strategies"""
        self.assertEqual(1, 1)

        shares = qt.filter_stock_codes(index='000300.SH', date='20210101')
        print(len(shares), shares[:10])

        alpha = IndexEnhancement()
        op = qt.Operator(alpha, signal_type='PT')

        op.op_type = 'stepwise'
        op.set_blender("0.8*s0", 'close')
        # op.run(mode=1,
        #        invest_start='20210101',
        #        invest_end='20220501',
        #        asset_type='E',
        #        invest_cash_amounts=[1000000],
        #        asset_pool=shares,
        #        trade_batch_size=100,
        #        sell_batch_size=1,
        #        trade_log=True)
        self.assertTrue(True)

    def test_grid_trading(self):
        """ test for strategy GridTrading"""
        self.assertEqual(1, 1)

        alpha = GridTrading()
        op = qt.Operator(alpha, signal_type='PT')

        op.op_type = 'batch'
        op.set_blender("1.0*s0", 'close')
        # op.run(
        #         mode=1,
        #         invest_start='20220401',
        #         invest_end='20220731',
        #         invest_cash_amounts=[1000000],
        #         asset_type='IDX',
        #         asset_pool=['000300.SH'],
        #         trade_batch_size=0,
        #         sell_batch_size=0,
        #         trade_log=True,
        #         allow_sell_short=True,
        # )
        self.assertTrue(True)

    def test_get_history_data(self):
        """
        """
        # extract normal data
        data = qt.get_history_data(htypes='open, high, low, close',
                                   start='20211230',
                                   end='20220110',
                                   shares='000001.SZ',
                                   freq='d',
                                   adj='n',
                                   asset_type='E')
        print(f'not adjusted data: \n{data}')
        # extract back adjusted price for one asset
        data = qt.get_history_data(htypes='open, high, low, close',
                                   start='20211230',
                                   end='20220110',
                                   shares='000001.SZ',
                                   freq='d',
                                   adj='b',
                                   asset_type='E')
        print(f'adjusted data: \n{data}')
        # extract back adjusted prices for both assets mixed
        data = qt.get_history_data(htypes='open, high, low, close',
                                   start='20211230',
                                   end='20220110',
                                   shares='000001.SZ, 512000.SH',
                                   freq='d',
                                   adj='b',
                                   asset_type='E, FD')
        print(f'adjusted data: \n{data}')
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()


# coding=utf-8
# ======================================
# File:     test_tui.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2024-04-09
# Desc:
#   Unittest for the TUI
# ======================================

import unittest

import qteasy.trader
from qteasy import Operator, DataSource
from qteasy.trader_tui import TraderApp
from qteasy.broker import SimulatorBroker


class TestTUI(unittest.TestCase):

    def setUp(self) -> None:
        """Set up test fixtures, if any."""
        print('Setting up test Trader...')
        operator = Operator(strategies=['macd', 'dma'], op_type='step')
        operator.set_parameter(
                stg_id='dma',
                window_length=20,
                strategy_run_freq='H'
        )
        operator.set_parameter(
                stg_id='macd',
                window_length=30,
                strategy_run_freq='30min',
        )
        broker = SimulatorBroker()
        config = {
            'mode':                  0,
            'time_zone':             'local',
            'market_open_time_am':   '09:30:00',
            'market_close_time_pm':  '15:30:00',
            'market_open_time_pm':   '13:00:00',
            'market_close_time_am':  '11:30:00',
            'exchange':              'SSE',
            'cash_delivery_period':  0,
            'stock_delivery_period': 0,
            'asset_pool':            '000001.SZ, 000002.SZ, 000004.SZ, 000005.SZ, 000006.SZ, 000007.SZ',
            'asset_type':            'E',
            'PT_buy_threshold':      0.05,
            'PT_sell_threshold':     0.05,
            'allow_sell_short':      False,
            'invest_start':          '2018-01-01',
            'opti_start':            '2018-01-01',
        }
        # 创建测试数据源
        data_test_dir = '../qteasy/data_test/'
        # 创建一个专用的测试数据源，以免与已有的文件混淆，不需要测试所有的数据源，因为相关测试在test_datasource中已经完成
        test_ds = DataSource('file', file_type='csv', file_loc=data_test_dir)

        self.trader = qteasy.trader.Trader(
                account_id=1,
                operator=operator,
                broker=broker,
                config=config,
                datasource=test_ds,
        )

    def test_action_toggle_dark(self):
        ''' test action_toggle_dark '''
        trader = self.trader
        app = TraderApp(trader=trader)
        app.action_toggle_dark()
        self.assertEqual(app.dark, False)
        app.action_toggle_dark()
        self.assertEqual(app.dark, True)


if __name__ == '__main__':
    # unittest.main()
    pass



# coding=utf-8
# ======================================
# File:     broker.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2023-04-09
# Desc:
#   class Broker for trader to submit
# trading orders and get trading results
# Broker classes are supposed to be
# inherited from Broker, representing
# different trading platforms or brokers
# ======================================

from queue import Queue
from abc import abstractmethod, ABCMeta

import numpy as np
import time

from qteasy import QT_CONFIG
from .utilfuncs import get_current_timezone_datetime

CASH_DECIMAL_PLACES = QT_CONFIG['cash_decimal_places']
AMOUNT_DECIMAL_PLACES = QT_CONFIG['amount_decimal_places']


def _verify_trade_result(trade_result, order_qty):
    """ 检查result，确认是否符合基本要求:
    包括trade_result各个组份的类型是否正确、数据是否超过范围

    Parameters:
    -----------
    trade_result: tuple: (result_type, qty, filled_price, fee)
        result_type: str
        qty: float
        filled_price: float
        fee: float
    order_qty: float
        订单的报价数量
    """
    result_type, qty, filled_price, fee = trade_result

    if not isinstance(result_type, str):
        raise TypeError(f'result_type should be str, but got {type(result_type)}')
    if result_type not in ['filled', 'partial-filled', 'canceled']:
        raise ValueError(f'result_type should be one of ["filled", "canceled"], but got {result_type}')
    if not isinstance(qty, (int, float)):
        raise TypeError(f'qty should be int or float, but got {type(qty)}')
    if qty <= 0:
        raise ValueError(f'qty should be greater than 0, but got {qty}')
    if qty > order_qty:
        raise ValueError(f'qty should be less than or equal to order["qty"], but got {qty}')
    if not isinstance(filled_price, (int, float)):
        raise TypeError(f'filled_price should be int or float, but got {type(filled_price)}')
    if filled_price < 0:
        raise ValueError(f'filled_price should be greater than 0, but got {filled_price}')
    if result_type == 'canceled' and filled_price != 0:
        raise ValueError(f'filled_price should be 0 when result_type is "canceled", but got {filled_price}')
    if not isinstance(fee, (int, float)):
        raise TypeError(f'fee should be int or float, but got {type(fee)}')
    if fee < 0:
        raise ValueError(f'fee should be greater than 0, but got {fee}')

    return True


class Broker(object):
    """ Broker是交易所的抽象，它接受交易订单并返回交易结果

    BaseBroker定义了Broker的基本接口，所有的Broker都必须继承自BaseBroker，
    以实现不同交易所的接口

    Attributes:
    -----------
    order_queue: Queue
        交易订单队列，每个交易订单都是一个list，包含多个交易订单
    result_queue: Queue
        交易结果队列，每个交易结果都是一个list，包含多个交易结果
    status: str
        Broker的状态，可以是 'init', 'running', 'stopped', 'paused'
        - init: Broker刚刚创建，还没有开始运行
        - running: Broker正在运行, 接受交易订单并返回交易结果
        - stopped: Broker已经停止运行, 停止所有操作，未完成的订单将被丢弃，并退出主循环
        - paused: Broker暂停运行，未完成的订单将被暂停，可以接受交易订单，但不会返回交易结果

    Methods:
    --------
    register(debug=False, **kwargs):
        交易所的初始化程序，可以
    run()
        Broker的主循环，从order_queue中获取交易订单并处理，获得交易结果并放入result_queue中
    transaction(symbol, order_qty, order_price, direction, position='long', order_type='market'): abstract method
        一个Generator方法。交易所处理交易订单并获取交易结果, 在子类中必须实现这个方法
        这个方法在接受交易订单后，会以generator形式返回订单的处理结果，直至订单处理完毕或出现错误
    """
    __metaclass__ = ABCMeta

    # TODO: for v1.1:
    #  重构Broker类，使Broker基类提供通用的接口，如log_in, log_out, run, transaction等
    #  并将具体的交易所实现放在子类中，如SimulatorBroker, SimpleBroker等
    #  重构后的Broker类应该是一个抽象类，不能直接实例化，只能通过子类实例化
    #  重构后的Broker类提供的接口可以通用于模拟交易所和真实交易所，同时所有的接口以真实交易所的接口
    #  为标准，模拟交易所的接口应该尽量与真实交易所的接口保持一致
    def __init__(self, data_source=None):
        """ 生成一个Broker对象

        Parameters
        ----------
        data_source: DataSource or None
            交易所的数据源，可以是数据库、API等，如果为None，则使用QT_DATA_SOURCE
        """
        if data_source is None:
            from qteasy import QT_DATA_SOURCE
            data_source = QT_DATA_SOURCE
        self._data_source = data_source
        self.broker_name = 'BaseBroker'
        self.user_name = ''
        self.password = ''
        self.token = ''
        self.status = 'init'  # init, running, stopped, paused
        self.debug = False
        self.is_registered = False

        self.time_zone = 'local'
        self.init_time = get_current_timezone_datetime(self.time_zone).strftime('%Y-%m-%d %H:%M:%S')

        self.order_queue = Queue()
        self.result_queue = Queue()
        self.broker_messages = Queue()

    @property
    def data_source(self):
        return self._data_source

    def register(self, debug=False, **kwargs):
        """ Broker对象在开始运行前的注册过程，作用是设置broker的状态为is_registered
        Override这个函数，以添加更多处理

        只有is_registered == True时，broker才能够运行

        Parameters
        ----------
        debug: bool, default: False
            是否进入debug mode
        kwargs:

        Return
        ------
        None
        """
        # override this function if necessary
        self.is_registered = True
        self.debug = debug

    def log_out_broker(self):
        """ Broker对象在关闭前的处理过程。
        Override这个函数，以添加更多处理
        """
        # override this function if necessary
        pass

    def run(self):
        """ Broker的主循环，从order_queue中获取交易订单并处理，获得交易结果并放入result_queue中
        order_queue中的每一个交易订单都由transaction函数来处理并获取交易结果，transaction函数是
        一个Generator，可以分批返回交易结果，直至订单处理完毕或出现错误。每一个transaction都会在单
        独的线程中运行
        """
        from threading import Thread
        if not self.is_registered:
            raise RuntimeError(f'broker is not registered!')
        if self.debug:
            self.send_message(f'is running...')
        self.status = 'init'
        while True:
            try:
                time.sleep(0.05)
                if self.status == 'stopped':
                    # 如果Broker正常退出，处理尚未提取的交易订单，这些订单将不会被处理，会提示用户取消订单
                    print(f'Stopping un-processed orders in broker...')
                    un_processed_orders = []
                    while not self.order_queue.empty():
                        un_processed_orders.append(self.order_queue.get())
                        self.order_queue.task_done()

                    if len(un_processed_orders) > 0:
                        print(f'Un-processed orders: {un_processed_orders}')

                    # 交易所关闭后的处理程序
                    self.log_out_broker()
                    print(f'Broker is stopped')
                    break

                # 如果Broker处于暂停状态，则不处理交易订单
                if self.status == 'paused':
                    continue

                # 如果order_queue为空，则不处理交易订单
                if self.order_queue.empty():
                    continue

                # order_queue不为空，提取交易订单，在一个单独的thread中调用self._get_result()处理交易订单
                order = self.order_queue.get()  # order is a dict, should be packed in a tuple
                t = Thread(target=self._get_result, args=(order, ), daemon=True)
                t.start()

            except KeyboardInterrupt:
                # 如果Broker被用户强制退出，处理尚未完成的交易订单
                if self.debug:
                    self.send_message('Broker will be stopped by user.')
                self.status = 'stopped'
                continue
            except Exception as e:
                # 如果Broker出现异常，打印异常信息，并继续运行
                self.send_message(f'Runtime exception: {e}, please check system log for details')
                if self.debug:
                    import traceback
                    traceback.print_exc()
                continue

        return

    def send_message(self, message: str, new_line=True):
        """ 将消息放入broker的消息队列
        """
        if self.debug:
            message = f'[DEBUG]-{message}'
        message = f'[{self.broker_name}]: {message}'
        if not new_line:
            message += '_R'
        self.broker_messages.put(message)

    def _submit_order(self, order):
        """

        :param order:
        :return:
        """
        pass

    def _parse_order(self, order):
        """ 解析交易订单，提取其关键信息，并将order的状态改为"submitted"

        Parameters:
        -----------
        order: dict 交易订单dict，详细信息如下:
            {'order_id': 订单ID,
             'pos_id': position ID,
             'direction',
             'order_type',
             'qty',
             'price',
             'submitted_time',
             'status'}

        Returns:
        --------
        order info: tuple(symbol, qty, price, direction, position)
        symbol:
        qty:
        price:
        direction:
        position:
        """

        if self.debug:
            self.send_message(f'_parse_order():\nsubmit order components of order(ID) {order["order_id"]}:\n'
                              f'quantity:{order["qty"]}\norder_price={order["price"]}\n'
                              f'order_direction={order["direction"]}\n')
        pos_id = order['pos_id']
        from qteasy.trade_recording import get_position_by_id
        try:
            position = get_position_by_id(pos_id=pos_id, data_source=self.data_source)
        except RuntimeError as e:
            raise RuntimeError(e)
        symbol = position['symbol']
        position = position['position']
        qty = order['qty']
        price = order['price']
        direction = order['direction']
        order_type = order['order_type']
        return order_type, symbol, qty, price, direction, position

    def _get_result(self, order):
        """ 解析订单信息，将关键信息提交给transaction，获取交易结果后，解析交易结果，将结果放入result_queue

        Parameters:
        -----------
        order: dict 交易订单dict，详细信息如下:
            {'order_id': 订单ID,
             'pos_id': position ID,
             'direction',
             'order_type',
             'qty',
             'price',
             'submitted_time',
             'status'}

        Returns:
        --------
        raw_trade_result: dict, 初步交易结果dict,详细信息如下：
            {'order_id',
             'filled_qty',
             'price',
             'transaction_fee',
             'execution_time',
             'canceled_qty',
             'delivery_amount',
             'delivery_status'
             }
        """

        order_type, symbol, qty, price, direction, position = self._parse_order(order)
        for trade_result in self.transaction(
                order_type=order_type,
                symbol=symbol,
                order_qty=qty,
                order_price=price,
                direction=direction,
                position=position
        ):
            result_type, qty, filled_price, fee = trade_result
            _verify_trade_result(trade_result, qty)

            # 确认数据格式正确后，将数据圆整到合适的精度，并组装为raw_trade_result
            if self.debug:
                self.send_message(f'method: _get_result(): got transaction result for order(ID) {order["order_id"]}\n'
                                  f'result_type={result_type}, \nqty={qty}, \n'
                                  f'filled_price={filled_price}, \nfee={fee}')
            # 圆整qty、filled_qty和fee
            qty = round(qty, AMOUNT_DECIMAL_PLACES)
            filled_price = round(filled_price, CASH_DECIMAL_PLACES)
            transaction_fee = round(fee, CASH_DECIMAL_PLACES)

            filled_qty = 0
            canceled_qty = 0
            if result_type in ['filled', 'partial-filled']:
                filled_qty = qty
            elif result_type == 'canceled':
                canceled_qty = qty
            else:
                raise ValueError(f'Unknown result_type: {result_type}, should be one of ["filled", "canceled"]')

            current_datetime = get_current_timezone_datetime(self.time_zone)
            raw_trade_result = {
                'order_id':        order['order_id'],
                'filled_qty':      filled_qty,
                'price':           filled_price,
                'transaction_fee': transaction_fee,
                'execution_time':  current_datetime.strftime('%Y-%m-%d %H:%M:%S'),
                'canceled_qty':    canceled_qty,
                'delivery_amount': 0,
                'delivery_status': 'ND',
            }

            # 将trade_result放入result_queue
            if self.debug:
                self.send_message(f'method _get_result(): created raw trade result for order(ID) {order["order_id"]}:\n'

                                  f'{raw_trade_result}')
            self.result_queue.put(raw_trade_result)

        # 全部订单处理完毕或发生错误后结束
        return


    @abstractmethod
    def transaction(self, symbol, order_qty, order_price, direction, position='long', order_type='market'):
        """ 交易所处理交易订单并获取交易结果, 抽象方法，需要由用户在子类中实现
        应该将transaction定义为Generator，分批完成交易，并分批返回

        Parameters:
        -----------
        symbol: str
            交易标的股票代码
        order_qty: float
            挂单数量
        order_price: float
            交易报价
        direction: str
            交易订单方向，'buy' 或者 'sell'
        position: str, default 'long'
            交易订单的持仓方向，'long' 或者 'short'
        order_type: str, default 'market'
            交易订单类型，'market' 或者 'limit'

        Returns / Yields:
        -----------------
        tuple: (result_type, qty, price, fee)
            result_type: str 交易结果类型:
                'filled' - 成交,
                'partial_filled' - 部分成交,
                'canceled' - 取消
                'failed' - 失败
            qty: float
                成交/取消数量，这个数字应该小于等于order_qty，且大于等于0
            price: float
                成交价格, 如果是取消交易，价格为0或任意数字
            fee: float
                交易费用，交易费用应该大于等于0

        Notes:
        ------
        1. 成交结果可以分多次返回，直至运行结束或交易失败/取消
        2. 如果多次返回成交结果，需要使用yield关键字将method定义为一个generator

        Examples:
        ---------
        >>> broker = Broker()
        >>> # 交易所返回完全成交结果
        >>> broker.transaction(100, 10, 'buy')
        ('filled', 100, 10, 5)
        >>> # 交易所返回部分成交结果
        >>> broker.transaction(100, 10, 'buy')
        ('partial_filled', 50, 10, 5)
        >>> # 交易所返回分批成交结果
        >>> broker.transaction(200, 10, 'buy')
        ('partial_filled', 100, 10, 5)
        ('filled', 100, 1)
        >>> # 交易所返回取消结果
        >>> broker.transaction(100, 10, 'buy')
        ('canceled', 100, 0, 0)
        >>> # 交易所返回失败结果
        >>> broker.transaction(100, 10, 'buy')
        ('failed', 0, 0, 0)
        """
        pass


class SimpleBroker(Broker):
    """ SimpleBroker接到交易订单后，立即返回交易结果
    交易结果总是完全成交，根据moq调整交易批量，根据设定的交易费率计算交易费用，滑点是按照百分比计算的

    Parameters
    ----------
    """

    def __init__(self, data_source=None):
        super(SimpleBroker, self).__init__(data_source=data_source)
        self.broker_name = 'SimpleBroker'

    def transaction(self, symbol, order_qty, order_price, direction, position='long', order_type='market'):
        """ 订单立即成交

        Parameters:
        ----------
        symbol: str
            挂单交易标的股票代码
        order_qty: float
            挂单数量
        order_price: float
            挂单报价
        direction: str
            交易订单方向，'buy' 或者 'sell'
        position: str, default 'long'
            交易订单的持仓方向，'long' 或者 'short'
        order_type: str, default 'market'
            交易订单类型，'market' 或者 'limit'

        Yields:
        -------
        trade_results: tuple: (result_type, qty, price, fee)
            result_type: str 交易结果类型:
                'filled' - 成交,
                'partial_filled' - 部分成交,
                'canceled' - 取消
                'failed' - 失败
            qty: float
                成交/取消数量，这个数字应该小于等于order_qty，且大于等于0
            price: float
                成交价格, 如果是取消交易，价格为0或任意数字
            fee: float
                交易费用，交易费用应该大于等于0
        """
        # TODO: consider, can these common codes be moved to base class?
        total_filled = 0

        while True:

            import time
            time.sleep(1)

            if total_filled >= order_qty:  # 订单完全成交，退出
                break

            if self.status == 'stopped':  # 当broker停止时，退出
                break

            if self.status == 'paused':  # 当broker暂停时，稍后重试
                continue

            qty = order_qty
            price = order_price
            fee: float = 5.
            trade_result = ('filled', qty, price, fee)
            yield trade_result

            total_filled += qty


class SimulatorBroker(Broker):
    """ QT 默认的模拟交易所，该交易所模拟真实的交易情况，从qt_config中读取交易费率、滑点等参数

    根据这些参数尽可能真实地模拟成交结果，特点如下：

    - 在收到交易订单后，根据订单类型和当前价格确认成交类型：
        - 如果订单类型是市价单：以当前价成交
        - 如果订单类型是限价单：若当前格低于叫买价，以当前价买入，若当前价高于叫卖价，以当前价卖出
    - 股票涨停时大概率买入交易失败，跌停时大概率卖出交易失败
    - 交易费率根据参数中的设置计算，包括固定费率、最低费用、滑点等
    """

    def __init__(self,
                 fee_rate_buy=0.0003,
                 fee_rate_sell=0.0001,
                 fee_min_buy=0.0,
                 fee_min_sell=0.0,
                 fee_fix_buy=0.0,
                 fee_fix_sell=0.0,
                 slipage=0.0,
                 moq_buy=0.0,
                 moq_sell=0.0,
                 delay=1.0,
                 price_deviation=0.0,
                 probabilities=(0.9, 0.08, 0.02),
                 data_source=None):
        """ 生成一个Broker对象

        Parameters
        ----------
        fee_rate_buy: float,
            买入操作的交易费率
        fee_rate_sell: float,
            卖出操作的交易费率
        fee_min_buy: float, default 0.0
            买入操作的最低交易费用
        fee_min_sell: float, default 0.0
            卖出操作的最低交易费用
        fee_fix_buy: float, default 0.0
            买入操作的固定交易费用，如果不为0，则忽略交易费率和最低费用
        fee_fix_sell: float, default 0.0
            卖出操作的固定交易费用，如果不为0，则忽略交易费率和最低费用
        slipage: float, default 0.0
            交易滑点, 当交易数量很大时，交易费用会被放大 slipage * (qty / 100) ** 2 倍
        moq_buy: float, default 0.0
            买入操作最小数量
        moq_sell: float, default 0.0
            卖出操作最小数量
        delay: float, default 1.0
            模拟交易延迟，单位为秒
        price_deviation: float, default 0.0
            模拟成交价格允许误差值。例如，当前价格100元，误差值为0.01，即允许价格误差为100*0.01 = 1元
            此时买入报价大于100-1元即可成交
            卖出报价小于100+1元即可成交
        probabilities: tuple of 3 floats, default (0.90, 0.08, 0.02)
            模拟完全成交、部分成交和未成交三种情况出现的概率

        """
        super(SimulatorBroker, self).__init__(data_source=data_source)
        self.broker_name = 'SimulatorBroker'
        self.fee_rate_buy = fee_rate_buy
        self.fee_rate_sell = fee_rate_sell
        self.fee_min_buy = fee_min_buy
        self.fee_min_sell = fee_min_sell
        self.fee_fix_buy = fee_fix_buy
        self.fee_fix_sell = fee_fix_sell
        self.slipage = slipage
        self.moq_buy = moq_buy
        self.moq_sell = moq_sell
        self.delay = delay
        self.price_deviation = price_deviation
        self.probabilities = probabilities

    def transaction(self, symbol, order_qty, order_price, direction, position='long', order_type='market'):
        """ 读取实时价格模拟成交结果
        """

        total_filled = 0

        while True:

            import time
            time.sleep(1)  # 延迟1秒重试

            if total_filled >= order_qty:  # 订单完全成交，退出
                break

            if self.status == 'stopped':  # 当broker停止时，退出
                break

            if self.status == 'paused':  # 当broker暂停时，稍后重试
                continue

            # 获取当前实时价格
            retry = 0
            from .emfuncs import stock_live_kline_price
            live_prices = stock_live_kline_price(symbol, freq='D', verbose=True, parallel=False)
            if not live_prices.empty:
                live_prices['close'] = live_prices['close'].astype('float')
                change = (live_prices['close'] / live_prices['pre_close'] - 1).iloc[-1]
                live_price = live_prices.close.iloc[-1]
                price_deviation = live_price * self.price_deviation
            else:  # 实时价格获取不成功，稍后重试
                self.send_message(f'live price of {symbol} can not be acquired at the moment, '
                                  f'order will not be processed, retries left: {retry}')
                retry -= 1
                live_price = 0
                change = 0
                price_deviation = 0
                if retry > 0:
                    continue
            # 重试超过max_retry，直接cancel订单
            if retry <= 0:
                self.send_message(f'order will be canceled because max retries exceeded')
                result_type = 'canceled'
            # 如果当前价高于挂卖价(允许误差由price_deviation控制)，大概率成交或部分成交
            if (live_price >= order_price - price_deviation) and (direction == 'sell'):
                result_type = np.random.choice(['filled', 'partial-filled', 'canceled'], p=self.probabilities)
                # 如果change非常接近-10%(跌停)，则非常大概率canceled
                if abs(change + 0.1) <= 0.001:
                    self.send_message(f'order will be probably canceled because -10% sell-limit')
                    result_type = np.random.choice(['filled', 'partial-filled', 'canceled'], p=(0.01, 0.01, 0.98))
            # 如果当前价低于挂买价(允许误差由price_deviation控制), 大概率成交或部分成交
            elif (live_price <= order_price + price_deviation) and (direction == 'buy'):
                result_type = np.random.choice(['filled', 'partial-filled', 'canceled'], p=self.probabilities)
                # 如果change非常接近+10%(涨停)，则非常大概率canceled
                if abs(change - 0.1) <= 0.001:
                    self.send_message(f'order will be probably canceled because +10% buy-limit')
                    result_type = np.random.choice(['filled', 'partial-filled', 'canceled'], p=(0.01, 0.01, 0.98))
            # 如果挂单类型为市价单且不受张跌停限制, 大概率成交或部分成交
            elif (order_type == 'market') and (-0.098 < change < 0.098):
                # 市价单，只要不涨停跌停即可市价成交
                result_type = np.random.choice(['filled', 'partial-filled', 'canceled'], p=self.probabilities)
            # 所有条件均不满足，稍后重试
            else:
                if self.debug:
                    self.send_message(f'current price does not satisfy quoted, retries left: {retry}')
                retry -= 1
                continue

            # 成交类型为成交或部分成交，计算成交数量及交易费用
            if result_type in ['filled', 'partial-filled']:
                # 部分成交及全部成交，计算成交数量
                filled_proportion = 1
                if result_type == 'partial-filled':
                    filled_proportion = np.random.choice([0.25, 0.5, 0.75], p=[0.3, 0.5, 0.2])  # 模拟交易所的部分成交
                remain_qty = order_qty - total_filled
                qty = remain_qty * filled_proportion
                if self.moq_buy > 0:
                    qty = np.trunc(qty / self.moq_buy) * self.moq_buy
                    if (qty < self.moq_buy) and (remain_qty > self.moq_buy):
                        qty = self.moq_buy  # 如果成交数量小于moq，但是剩余数量大于moq，那么成交数量就是moq
                    elif (qty < self.moq_buy) and (remain_qty <= self.moq_buy):
                        qty = remain_qty  # 如果成交数量小于moq，且剩余数量也小于moq，那么成交数量就是剩余数量
                        result_type = 'filled'
                # 如果确认成交，成交价格为当前价格
                order_price = live_price
                # 计算交易费用，根据买入/卖出费率计算
                if direction == 'buy':
                    transaction_fee = \
                        max(qty * order_price * self.fee_rate_buy, self.fee_min_buy) \
                            if self.fee_fix_buy == 0 \
                            else self.fee_fix_buy
                elif direction == 'sell':
                    transaction_fee = \
                        max(qty * order_price * self.fee_rate_sell, self.fee_min_sell) \
                            if self.fee_fix_sell == 0 \
                            else self.fee_min_sell
                else:
                    raise RuntimeError(f'invalid direction: {direction}')
                # 模拟交易滑点, 交易数量越大，对交易费用产生的影响越大
                if self.slipage > 0:
                    transaction_fee *= (1 + self.slipage * (qty / 100) ** 2)

            # 成交类型为取消，成交数量为0，成交价格为0，交易费用为0
            elif result_type in ['canceled']:  # result_type == 'canceled'
                transaction_fee = 0
                order_price = 0
                qty = order_qty - total_filled
            # 不能成交，稍后重试
            else:
                continue

            order_result = (result_type, qty, order_price, transaction_fee)
            yield order_result

            total_filled += qty


class NotImplementedBroker(Broker):
    """ NotImplementedBroker raises NotImplementedError when __init__() is called
    """

    def __init__(self, data_source=None):
        super(NotImplementedBroker, self).__init__(data_source=data_source)
        raise NotImplementedError('NotImplementedBroker is not implemented yet')

    def transaction(self, symbol, order_qty, order_price, direction, position='long', order_type='market'):
        pass


def get_broker(name: str = 'simulator', params=None):
    """ get broker object by broker name

    Parameters
    ----------
    name:str default: simulator
        the broker name
    params: dict or None
        the parameters of broker

    Return
    ------
    Broker
    """
    all_brokers = {
        'random':    SimulatorBroker,
        'simple':    SimpleBroker,
        'manual':    NotImplementedBroker,
        'simulator': SimulatorBroker,
    }
    names_to_be_deprecated = {'random': 'simulator'}

    if not isinstance(name, str):
        raise TypeError(f'name must be a string, got {type(name)}')

    if name in names_to_be_deprecated:
        import warnings
        warnings.warn(f'the broker {name} will be deprecated in next version, '
                      f'use {names_to_be_deprecated[name]} instead')
    broker_func = all_brokers.get(name, SimulatorBroker)
    if params is not None:
        return broker_func(**params)
    return broker_func()


# coding=utf-8
# ======================================
# Package:  qteasy
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-11
# Desc:
# QTEASY:
#  A fast and easy-to-use quant-investment
#  strategy research tool kit.
# ======================================

import os
import sys
import warnings

import tushare as ts
import numpy as np
import logging
from logging.handlers import TimedRotatingFileHandler
from argparse import Namespace

from .utilfuncs import is_float_like, is_integer_like
from .core import run, set_config, get_configurations, get_config, view_config_files
from .core import info, is_ready, configure, configuration, save_config, load_config, reset_config
from .core import get_basic_info, get_stock_info, get_data_overview, refill_data_source
from .core import get_history_data, filter_stock_codes, filter_stocks
from .core import reconnect_ds, get_table_info, get_table_overview, live_trade_accounts
from .history import HistoryPanel
from .history import dataframe_to_hp, stack_dataframes
from .qt_operator import Operator
from .strategy import BaseStrategy, RuleIterator, GeneralStg, FactorSorter
from .built_in import built_ins, built_in_list, built_in_strategies, get_built_in_strategy, built_in_doc
from .visual import candle
from .finance import CashPlan, set_cost, update_cost
from .database import DataSource, find_history_data
from ._arg_validators import QT_CONFIG, ConfigDict
from .trade_recording import delete_account


# qteasy版本信息
__version__ = '1.3.0'
version_info = Namespace(
        major=1,
        minor=3,
        patch=0,
        short=(1, 3),
        full=(1, 3, 0),
        string='1.3.0',
        tuple=('1', '3', '0'),
        releaselevel='beta',
)

# 解析qteasy的本地安装路径
QT_ROOT_PATH = os.path.normpath(os.path.join(os.path.dirname(__file__), os.pardir))
QT_ROOT_PATH = os.path.join(QT_ROOT_PATH, 'qteasy/')

# 准备从本地配置文件中读取预先存储的qteasy配置
qt_local_configs = {}

QT_CONFIG_FILE_INTRO = '# qteasy configuration file\n' \
                       '# following configurations will be loaded when initialize qteasy\n\n' \
                       '# example:\n' \
                       '# local_data_source = database\n\n'
# 读取configurations文件内容到config_lines列表中，如果文件不存在，则创建一个空文本文件
try:
    with open(os.path.join(QT_ROOT_PATH, 'qteasy.cfg')) as f:
        config_lines = f.readlines()

except FileNotFoundError as e:
    qt_config_file_path_name = os.path.join(QT_ROOT_PATH, 'qteasy.cfg')
    with open(qt_config_file_path_name, mode='w', encoding='utf-8') as f:
        intro = QT_CONFIG_FILE_INTRO
        f.write(intro)

    config_lines = []  # 本地配置文件行
    msg = f'qteasy.cfg not found, a new configuration file is created, \nview file at: {qt_config_file_path_name}'
    warnings.warn(msg)
except Exception as e:
    msg = f'Error reading configuration file, all configurations will fall back to default! \n{e}'
    warnings.warn(msg)
    config_lines = []

# 解析config_lines列表，依次读取所有存储的属性，所有属性存储的方式为：
# config_key = value
for line in config_lines:
    if line[0] == '#':  # 忽略注释行
        continue
    line = line.split('=')
    if len(line) == 2:
        arg_name = line[0].strip()
        read_value = line[1].strip()
        if (read_value[0] in ['\'', '"']) and (read_value[-1] == ['\'', '"']):
            read_value = str(read_value[1:-1])
        elif read_value == 'True':
            read_value = True
        elif read_value == 'False':
            read_value = False
        elif read_value == 'None':
            read_value = None
        elif is_integer_like(read_value):
            read_value = int(read_value)
        elif is_float_like(read_value):
            read_value = float(read_value)
        else:
            pass

        arg_value = read_value
        try:
            qt_local_configs[arg_name] = arg_value
        except Exception as e:
            msg = f'{e}, invalid parameter: {arg_name}'
            warnings.warn(msg)

# 读取tushare token，如果读取失败，抛出warning
try:
    TUSHARE_TOKEN = qt_local_configs['tushare_token']
    ts.set_token(TUSHARE_TOKEN)
except Exception as e:
    msg = f'Failed Loading tushare_token, configure it in qteasy.cfg:\n' \
            f'tushare_token = your_token\n' \
            f'for more information, check qteasy tutorial: ' \
            f'https://qteasy.readthedocs.io/zh/latest/tutorials/1-get-started.html'
    warnings.warn(msg)

# 读取其他本地配置属性，更新QT_CONFIG, 允许用户自定义参数存在
configure(only_built_in_keys=False, **qt_local_configs)

# 连接默认的本地数据源
QT_DATA_SOURCE = DataSource(
        source_type=QT_CONFIG['local_data_source'],
        file_type=QT_CONFIG['local_data_file_type'],
        file_loc=QT_CONFIG['local_data_file_path'],
        host=QT_CONFIG['local_db_host'],
        port=QT_CONFIG['local_db_port'],
        user=QT_CONFIG['local_db_user'],
        password=QT_CONFIG['local_db_password'],
        db_name=QT_CONFIG['local_db_name']
)

# 初始化默认交易日历
QT_TRADE_CALENDAR = QT_DATA_SOURCE.read_table_data('trade_calendar')
if not QT_TRADE_CALENDAR.empty:
    QT_TRADE_CALENDAR = QT_TRADE_CALENDAR
else:
    QT_TRADE_CALENDAR = None
    msg = 'trade calendar is not loaded, some utility functions may not work properly, ' \
            'to download trade calendar, run \n"qt.refill_data_source(tables=\'trade_calendar\')"'
    warnings.warn(msg)

# 设置qteasy运行过程中忽略某些numpy计算错误报警
np.seterr(divide='ignore', invalid='ignore')

# 设置qteasy回测交易报告以及错误报告的存储路径
QT_SYS_LOG_PATH = os.path.join(QT_ROOT_PATH, QT_CONFIG['sys_log_file_path'])  # 系统日志存储路径
QT_TRADE_LOG_PATH = os.path.join(QT_ROOT_PATH, QT_CONFIG['trade_log_file_path'])  # 交易记录存储路径，包括回测和实盘交易

# 设置系统日志以及交易日志的存储路径，如果路径不存在，则新建一个文件夹
os.makedirs(QT_SYS_LOG_PATH, exist_ok=True)
os.makedirs(QT_TRADE_LOG_PATH, exist_ok=True)
# 设置loggings，创建logger
debug_handler = logging.handlers.TimedRotatingFileHandler(filename=os.path.join(QT_SYS_LOG_PATH, 'qteasy.log'),
                                                          backupCount=3, when='midnight')
error_handler = logging.StreamHandler()
debug_handler.setLevel(logging.DEBUG)
error_handler.setLevel(logging.ERROR)
formatter = logging.Formatter('[%(asctime)s]:%(levelname)s - %(module)s -: %(message)s')
debug_handler.setFormatter(formatter)
error_handler.setFormatter(formatter)
# 创建两个logger，一个是正常运行(回测、优化等）的logger，另一个是实盘运行logger
logger_core = logging.getLogger('core')
logger_core.addHandler(debug_handler)
logger_core.addHandler(error_handler)
logger_core.setLevel(logging.INFO)
logger_core.propagate = False

logger_core.info('qteasy loaded!')

# 检查当前python的版本信息，打印相关的warning信息
py_version = sys.version_info
py_ver_major = py_version.major
py_ver_minor = py_version.minor

# if py_ver_major < 3 or (py_ver_major == 3 and py_ver_minor <= 6):
#     warnings.warn(f'qteasy is not fully tested on Python version {py_ver_major}.{py_ver_minor}, '
#                   f'it is highly recommended to use python 3.8 or later for qteasy!')
# elif py_ver_major == 3 and py_ver_minor == 7:
#     warnings.warn(f'A few features of qteasy are planned to be deprecated in Python 3.7, '
#                   f'please updated to python 3.8 or later for qteasy!', DeprecationWarning)


__all__ = [
    'run', 'set_config', 'get_configurations', 'get_config', 'view_config_files',
    'info', 'is_ready', 'configure', 'configuration', 'save_config', 'load_config', 'reset_config',
    'get_basic_info', 'get_stock_info', 'get_data_overview', 'refill_data_source',
    'get_history_data', 'filter_stock_codes', 'filter_stocks',
    'reconnect_ds', 'get_table_info', 'get_table_overview',
    'HistoryPanel', 'dataframe_to_hp', 'stack_dataframes',
    'Operator',
    'BaseStrategy', 'RuleIterator', 'GeneralStg', 'FactorSorter',
    'built_ins', 'built_in_list', 'built_in_strategies', 'get_built_in_strategy',
    'candle',
    'CashPlan', 'set_cost', 'update_cost',
    'DataSource', 'find_history_data',
    'QT_TRADE_CALENDAR', 'QT_TRADE_LOG_PATH', 'QT_ROOT_PATH', 'QT_SYS_LOG_PATH',
    'QT_DATA_SOURCE', 'QT_CONFIG_FILE_INTRO',
    'utilfuncs',
    'QT_CONFIG', 'ConfigDict', '__version__', 'version_info',
    'logger_core', 'live_trade_accounts', 'delete_account',
]
